import string
import time
from typing import Iterable, List, Optional, Tuple

from .. import procfile, procrec, proctree, utility
from . import CommandError


__all__ = 'run',


def _record_loop(
    interval: float,
    flat_node_list: list,
    recnum: Optional[int] = None,
    stop_without_result: bool = False,
) -> Iterable[Tuple[float, int]]:
    count = 1
    while True:
        start = time.time()

        yield start, count

        count += 1
        if recnum and count > recnum or stop_without_result and not flat_node_list:
            break

        latency = time.time() - start
        time.sleep(max(0, interval - latency))


def run(
    procfile_list: List[str],
    database_file: str,
    interval: float,
    procfs: str,
    environment: Optional[list] = None,
    query: Optional[str] = None,
    pid_list: Optional[str] = None,
    recnum: Optional[int] = None,
    reevalnum: Optional[int] = None,
    stop_without_result: bool = False,
):
    readers = {k: v for k, v in procfile.registry.items() if k in procfile_list}
    forest = proctree.Forest(readers, procfs)

    query_tpl = string.Template(query)
    pid_list_tpl = string.Template(pid_list)
    meta = utility.get_meta(procfile_list, procfs)
    flat_node_list = []  # do not re-bind, it's passed by reference to _record_loop
    with procrec.SqliteStorage(database_file, procfile_list, meta) as store:
        for ts, count in _record_loop(interval, flat_node_list, recnum, stop_without_result):
            if environment and (count == 1 or reevalnum and (count + 1) % reevalnum == 0):
                env_dict = utility.evaluate(environment)
                if query:
                    query = query_tpl.safe_substitute(env_dict)
                if pid_list:
                    pid_list = pid_list_tpl.safe_substitute(env_dict)

            branch_pids = [int(p) for p in pid_list.split(',') if p] if pid_list else None
            result = forest.get_roots(branch_pids)

            if query:
                try:
                    result = proctree.query(result, query)
                except proctree.JsonPathQueryError as ex:
                    raise CommandError(str(ex)) from ex

            flat_node_list[:] = proctree.flatten(result, procfile_list)
            store.record(ts, flat_node_list)
